from typing import List, Optional
import sqlalchemy as sa
import sqlalchemy.orm as orm
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.orm import Session
from importlib import import_module

__factory = None


def db_init(database_uri: str, database: DeclarativeMeta, check_same_thread: Optional[bool] = False,
            echo=False) -> None:
    """Create session factory. Use before creating the first session"""

    global __factory

    if __factory:
        return

    conn_str = f'{database_uri}?check_same_thread={check_same_thread}' if isinstance(check_same_thread,
                                                                                     bool) else database_uri

    engine = sa.create_engine(conn_str, echo=echo)
    __factory = orm.sessionmaker(bind=engine)

    database.metadata.create_all(engine)


def create_session() -> Session:
    return __factory()


def add_urls(urlpatterns: List[str]) -> None:
    for file in urlpatterns:
        import_module(file)



"""Django structure for your flask project.

To start project use command

$ flask-dj startproject project

"""


from .__project_info__ import __version__, __title__, __description__, \
    __author_email__, __author__, __keywords__, __license__
from .project_creator import startproject, command
from . import manage
from . import urls



from setuptools import setup
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'PyPI_README.md'), encoding='utf-8') as f:
    long_description = f.read()

about = {}
with open(path.join(here, 'flask_dj', '__project_info__.py')) as f:
    exec(f.read(), about)

packages = ["flask_dj"]
requirements = [
        'waitress>=1.4.3',
        'flask>=1.1.1',
        'Flask-Migrate>=2.5.2',
        'Flask-Script>=2.0.6',
        'Flask-SQLAlchemy>=2.4.1',
        'WTForms>=2.2.1',
        'SQLAlchemy>=1.3.13',
        'Flask-WTF>=0.14.3',
        'Flask-Login>=0.5.0'
    ]

setup(
    name=about["__title__"],  # Required
    version=about["__version__"],  # Required
    description=about["__description__"],  # Optional
    long_description=long_description,  # Optional
    long_description_content_type='text/markdown',  # Optional (see note above)
    url='https://github.com/AlexandrovRoman/Flask-DJ',  # Optional
    author=about["__author__"],  # Optional
    author_email=about["__author_email__"],  # Optional
    classifiers=[  # Optional
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        f'License :: OSI Approved :: {about["__license__"]}',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    keywords=about["__keywords__"],  # Optional
    packages=packages,  # Required
    install_requires=requirements,
    entry_points={
        'console_scripts': [
            'flask-dj = flask_dj:command',
        ],
    },
    project_urls={
        "Source": "https://github.com/AlexandrovRoman/Flask-DJ",
        "Docs": "https://flask-dj.readthedocs.io/en/latest/"
    }
)



from argparse import ArgumentParser
from typing import Iterable
from flask_dj.templates import config_file, init_file, urls_file, manage_file, utils_urls
from string import ascii_letters
from random import choices, randint
from sys import argv
from os.path import join
from ._utils import create_file, create_folder, valid_folder_name


class ProjectConstructor:
    """
    Your project creator
    :param project_name: str
    :param path: str
    :param need_templates: bool - Do you need to create the templates folder
    :param need_static: bool - Do you need to create the static folder
    """
    def __init__(self, project_name=None, path='', need_templates=False, need_static=False):
        self.path = path
        self.project_name = project_name
        valid_folder_name(project_name)
        self.project_path = join(self.path, self.project_name) if self.path else self.project_name
        self.main_app_path = join(self.project_path, self.project_name)
        self.need_templates = need_templates
        self.need_static = need_static

    def startproject(self):
        self._create_project_folder()
        self._create_main()
        self._create_manage()
        self._create_utils()
        if self.need_templates:
            self._create_templates()
        if self.need_static:
            self._create_static()

    def _create_project_folder(self):
        create_folder(self.project_path)

    def _create_main(self):
        self._create_main_app_folder()
        self._create_init()
        self._create_config()
        self._create_urls()

    def _create_main_app_folder(self):
        create_folder(self.main_app_path)

    def _create_init(self):
        create_file(self.main_app_path, '__init__', init_file)

    def _create_config(self):
        create_file(self.main_app_path, 'config',
                    config_file.format(project_name=self.project_name, csrf=self.generate_key(randint(9, 12)),
                                       secret_key=self.generate_key(randint(9, 12))))

    @staticmethod
    def generate_key(length: int):
        return "".join(choices(ascii_letters, k=length))

    def _create_urls(self):
        create_file(self.main_app_path, 'urls', urls_file.format(functions="add_relative_path, include"))

    def _create_manage(self):
        create_file(self.project_path, 'manage',
                    manage_file.format(project_name=self.project_name))

    def _create_utils(self):
        self._create_utils_folder()
        self._create_utils_urls()

    def _create_utils_folder(self):
        create_folder(join(self.project_path, 'utils'))

    def _create_utils_urls(self):
        create_file(join(self.project_path, 'utils'), 'urls', utils_urls.format(project_name=self.project_name))

    def _create_templates(self):
        create_folder(join(self.project_path, 'templates'))

    def _create_static(self):
        create_folder(join(self.project_path, 'static'))


def command():
    commands = {
        'startproject': console_creation
    }

    try:
        commands[argv[1]]()
    except KeyError:
        print("Input correct command")
        print_commands(commands)
    except IndexError:
        print("Input command")
        print_commands(commands)


def console_creation():
    parser = ArgumentParser()
    parser.add_argument("command")
    parser.add_argument("project_name")
    parser.add_argument("--templates", "-t", action="store_true", default=False)
    parser.add_argument("--static", "-st", action="store_true", default=False)
    args = parser.parse_args()

    startproject(args.project_name, need_templates=args.templates, need_static=args.static)


def startproject(project_name, path="", need_templates=False, need_static=False):
    ProjectConstructor(project_name, path, need_templates, need_static).startproject()


def print_commands(commands: Iterable):
    for key in commands:
        print(f"\t\t{key}")



from os.path import join, split
from os import makedirs, getcwd
from string import ascii_letters, digits
from flask_dj._exceptions import CreationError


def create_file(path: str, name: str, template: str) -> None:
    with open(join(path, f'{name}.py'), 'w') as f:
        f.write(template)


def create_folder(path: str) -> None:
    try:
        makedirs(path)
    except FileExistsError:
        raise CreationError('The directory with that name already exists.')
    except PermissionError:
        raise CreationError('You do not have enough permissions to act in this directory, '
                            'try changing users and try again.')


def get_project_name() -> str:
    return split(getcwd())[-1]


def valid_folder_name(name: str) -> None:
    if set(name).issubset(set(ascii_letters + digits + "_")) and name[0] not in digits:
        return
    raise ValueError("Invalid name")



from typing import NoReturn
from flask import Flask
from flask_migrate import MigrateCommand
from flask_script import Manager
from importlib import import_module
from flask_dj.templates import views_file, models_file, urls_file, forms_file
from waitress import serve
from os.path import join
from ._utils import get_project_name, create_folder, create_file, valid_folder_name

"""database-methods: https://flask-migrate.readthedocs.io/en/latest/"""
manager = None
app_ = None

commands = [

]


def init_manage_and_app(app: Flask):
    global manager, app_
    app_ = app
    manager = Manager(app)


def init_db_commands(models=[]):
    manager.add_command('db', MigrateCommand)

    for file in models:
        import_module(file)


def runserver(host: str, port: int) -> NoReturn:
    app_.run(host=host, port=port) if app_.debug else serve(app_, host=host, port=port)


def startapp(templates: bool, static: bool, name: str) -> None:
    """Create folder containing forms.py, models.py, urls.py, views.py"""
    valid_folder_name(name)

    create_folder(name)
    create_app_files(name)

    if templates:
        create_app_templates(name)
    if static:
        create_app_static(name)

    print(f'app {name} created')


def create_app_files(app_name: str) -> None:
    """Create .py files for your app"""
    project_name = get_project_name()

    create_file(app_name, 'views', views_file)
    create_file(app_name, 'models', models_file.format(project_name=project_name))
    create_file(app_name, 'urls', urls_file.format(functions="relative_path"))
    create_file(app_name, 'forms', forms_file)


def create_app_templates(app_name: str) -> None:
    create_folder(join("templates", app_name))


def create_app_static(app_name: str) -> None:
    create_folder(join("static", app_name))
